==> ec2init-0.11-py3-none-any.whl: ec2init/execution/base_async_procedure.py <==
import abc
from ec2init.execution.async_procedure import AsyncProcedure
from ec2init.execution.procedure import Procedure

from concurrent.futures import CancelledError, Executor, Future
from typing import List, Any
import logging

logger = logging.getLogger(__name__)

class BaseAsyncProcedure(AsyncProcedure):

    def execute(self) -> List['Procedure']:
        if not self.is_running():
            logger.debug("%s execute %d tasks with %d workers", self.name, len(self.tasks), self.max_workers)
            self._pool = self.create_pool()
            self._futures = [ (self.submit_task(self._pool, task), task, 1) for task in self.tasks]
        else:
            logger.debug("%s procedure is already running...", self.name)
        return []
    
    def wait_to_finish(self) -> None:
        if hasattr(self, "_pool") and hasattr(self, "_futures"):
            while True:
                to_remove = []
                for index, (f, task, retries) in enumerate(self._futures):
                    try:
                        f.result()
                    except CancelledError:
                        pass
                    except Exception:
                        if self._check_unhandled_exception(index, f, task, retries):
                            to_remove.append((f, task, retries))
                
                for f in to_remove:
                    self._futures.remove(f)

                if all([f.done() for f,_,_ in self._futures]):
                    logger.debug("%s finished all tasks", self.name)
                    break

            self._shutdown()
            self.completed_callback()
    
    def cancel(self) -> None:
        if hasattr(self, "_pool") and hasattr(self, "_futures"):
            for f,_,_  in self._futures:
                f.cancel()
            self._shutdown()

    def _shutdown(self) -> None:
        self._pool.shutdown()
        self.post_pool_shutdown()

    def check_unhandled_exception(self) -> None:
        if hasattr(self, "_futures"):
            to_remove = []
            for index, (f, task, retries) in enumerate(self._futures):
                if self._check_unhandled_exception(index, f, task, retries):
                    to_remove.append((f, task, retries))
            for f in to_remove:
                self._futures.remove(f)
    
    def _check_unhandled_exception(self, index: int, f :Future , task: Any, retries: int) -> bool:
        if f.done() and not f.cancelled() and f.exception() is not None:
            logger.warning("%s encountered an exception while processing task (%s) - Error: %s", self.name, str(task), str(f.exception()))
            self.task_rollback_function(task)
            if self.fail_plan or retries >= self.max_retries:
                logging.error("%s reached maximum number of retries, throwing an error (%s)", self.name, str(f.exception()))
                raise Exception(f.exception())
            else:
                logger.debug("%s retry failing task (%s) (%d/%d)", self.name, str(task), (retries+1), self.max_retries)
                self._futures.append((self.submit_task(self._pool, task), task, retries+1))
                return True 
        return False


    @abc.abstractmethod
    def task_rollback_function(self, task: Any) -> None:
        pass

    @abc.abstractmethod
    def create_pool(self) -> Executor:
        pass
    
    @abc.abstractmethod
    def submit_task(self, pool :Executor , task: Any):
        pass 
    
    @abc.abstractmethod
    def post_pool_shutdown(self) -> None:
        pass

==> ec2init-0.19.1-py3-none-any.whl: ec2init/execution/base_async_procedure.py <==
import abc
from ec2init.execution.async_procedure import AsyncProcedure
from ec2init.execution.procedure import Procedure

from concurrent.futures import CancelledError, Executor, Future
from typing import List, Any
import logging
from ec2init.utils.exceptions import ConfigException

logger = logging.getLogger(__name__)

class BaseAsyncProcedure(AsyncProcedure):

    def execute(self) -> List['Procedure']:
        if not self.is_running():
            logger.debug("%s execute %d tasks with %d workers", self.name, len(self.tasks), self.max_workers)
            self._pool = self.create_pool()
            self._futures = [ (self.submit_task(self._pool, task), task, 1) for task in self.tasks]
        else:
            logger.debug("%s procedure is already running...", self.name)
        return []
    
    def wait_to_finish(self) -> None:
        if hasattr(self, "_pool") and hasattr(self, "_futures"):
            while True:
                to_remove = []
                for index, (f, task, retries) in enumerate(self._futures):
                    try:
                        f.result()
                    except CancelledError:
                        pass
                    except ConfigException:
                        raise
                    except Exception:
                        if self._check_unhandled_exception(index, f, task, retries):
                            to_remove.append((f, task, retries))
                
                for f in to_remove:
                    self._futures.remove(f)

                if all([f.done() for f,_,_ in self._futures]):
                    logger.debug("%s finished all tasks", self.name)
                    break

            self._shutdown()
            self.completed_callback()
    
    def cancel(self) -> None:
        if hasattr(self, "_pool") and hasattr(self, "_futures"):
            for f,_,_  in self._futures:
                f.cancel()
            self._shutdown()

    def _shutdown(self) -> None:
        self._pool.shutdown()
        self.post_pool_shutdown()

    def check_unhandled_exception(self) -> None:
        if hasattr(self, "_futures"):
            to_remove = []
            for index, (f, task, retries) in enumerate(self._futures):
                if self._check_unhandled_exception(index, f, task, retries):
                    to_remove.append((f, task, retries))
            for f in to_remove:
                self._futures.remove(f)
    
    def _check_unhandled_exception(self, index: int, f :Future , task: Any, retries: int) -> bool:
        if f.done() and not f.cancelled() and f.exception() is not None:
            logger.warning("%s encountered an exception while processing task (%s) - Error: %s", self.name, str(task), str(f.exception()), exc_info=f.exception())
            self.task_rollback_function(task)
            if self.fail_plan or retries >= self.max_retries:
                logging.error("%s reached maximum number of retries, throwing an error (%s)", self.name, str(f.exception()))
                raise Exception(f.exception())
            else:
                logger.debug("%s retry failing task (%s) (%d/%d)", self.name, str(task), (retries+1), self.max_retries)
                self._futures.append((self.submit_task(self._pool, task), task, retries+1))
                return True 
        return False


    @abc.abstractmethod
    def task_rollback_function(self, task: Any) -> None:
        pass

    @abc.abstractmethod
    def create_pool(self) -> Executor:
        pass
    
    @abc.abstractmethod
    def submit_task(self, pool :Executor , task: Any):
        pass 
    
    @abc.abstractmethod
    def post_pool_shutdown(self) -> None:
        pass
